unit hiCriticalSection;

interface

uses
  Windows, Kol, Share, Debug;

type

  TSharedObject = class
    private
      FRefCount: Integer;
    protected
      property RefCount: Integer read FRefCount;
    public
      function IncRef: Integer;
      function DecRef: Integer;
  end;
  
  TNamedCriticalSection = class(TSharedObject)
    private
      
    public
      Name: string;
      CS: TRTLCriticalSection;
      
      constructor Create(const sName: string);
      destructor Destroy; override;
      
      procedure EnterCS;
      function TryEnterCS: Boolean;
      procedure LeaveCS;
  end;
  
  ThiCriticalSection = class(TDebug)
    private
      FName: string;
      FOurSection: TNamedCriticalSection;
      procedure SetName(Value: string);
      
    public
      _prop_WaitMode: Integer;
      
      _event_onEnter: THI_Event;
      _event_onBusy: THI_Event;
      
      property _prop_Name: string read FName write SetName;
      
      destructor Destroy; override;
      procedure _work_doEnterCS(var _Data: TData; Index: Word);
      procedure _work_doName(var _Data: TData; Index: Word);
  end;
  
  function TryEnterCriticalSection(var lpCriticalSection: TRTLCriticalSection): BOOL; stdcall;
           external 'kernel32.dll' name 'TryEnterCriticalSection';

implementation

var
  SectionsList: PList;
  ListCS: TRTLCriticalSection; //      

function FindSectionByName(const sName: string): TNamedCriticalSection;
var
  I: Integer;
begin
  for I := 0 to SectionsList.Count - 1 do
  begin
    Result := TNamedCriticalSection(SectionsList.Items[I]);
    if Result.Name = sName then Exit;
  end;
  Result := nil;
end;

//       .
//    - .
function HoldSection(const sName: string): TNamedCriticalSection;
begin
  EnterCriticalSection(ListCS);
    Result := FindSectionByName(sName);
    if Result = nil then
    begin
      Result := TNamedCriticalSection.Create(sName);
      SectionsList.Add(Result);
    end
    else
      Result.IncRef;
  LeaveCriticalSection(ListCS);
end;

//       
procedure ReleaseSection(Section: TNamedCriticalSection);
var
  C: Integer;
begin
  if Section = nil then Exit;
  EnterCriticalSection(ListCS);
    C := Section.DecRef; //      -   
    //    -   
    if C < 0 then SectionsList.Remove(Section);
  LeaveCriticalSection(ListCS);
end;

procedure ClearSections;
var
  I: Integer;
begin
  EnterCriticalSection(ListCS);
    for I := 0 to SectionsList.Count - 1 do
    begin
      TNamedCriticalSection(SectionsList.Items[I]).DecRef;
    end;
    SectionsList.Clear;
  LeaveCriticalSection(ListCS);
end;


// ====== TSharedObject ====== //

function TSharedObject.IncRef: Integer;
begin
  Result := InterlockedIncrement(FRefCount);
end;

function TSharedObject.DecRef: Integer;
begin
  Result := InterlockedDecrement(FRefCount);
  //        (-1) -
  //    
  if Result = -1 then Destroy;
end;



// ====== TNamedCriticalSection ====== //

constructor TNamedCriticalSection.Create(const sName: string);
begin
  inherited Create;
  Name := sName;
  InitializeCriticalSection(CS);
end;

destructor TNamedCriticalSection.Destroy;
begin
  DeleteCriticalSection(CS);
  inherited;
end; 

procedure TNamedCriticalSection.EnterCS;
begin
  EnterCriticalSection(CS);
end;

function TNamedCriticalSection.TryEnterCS: Boolean;
begin
  Result := TryEnterCriticalSection(CS);
end;

procedure TNamedCriticalSection.LeaveCS;
begin
  LeaveCriticalSection(CS);
end;


// ====== ThiCriticalSection ====== //

destructor ThiCriticalSection.Destroy;
begin
  ReleaseSection(FOurSection);
  inherited;
end;

procedure ThiCriticalSection.SetName(Value: string);
begin
  Value := LowerCase(Value);
  if (Value = '') or (Value = FName) then Exit;
  
  //     
  ReleaseSection(FOurSection);
  FName := Value;
  //     
  FOurSection := HoldSection(FName);
end;

procedure ThiCriticalSection._work_doEnterCS;
begin
  if FOurSection = nil then Exit;
  if _prop_WaitMode = 0 then // Wait
  begin
    FOurSection.EnterCS;
      _hi_OnEvent(_event_onEnter, _Data);
    FOurSection.LeaveCS;
  end
  else // No wait
    if FOurSection.TryEnterCS then 
    begin
      _hi_OnEvent(_event_onEnter, _Data);
      FOurSection.LeaveCS;
    end
    else
      _hi_CreateEvent(_Data, @_event_onBusy);
end;

procedure ThiCriticalSection._work_doName;
begin
  SetName(ToString(_Data));
end;


initialization
  SectionsList := NewList;
  InitializeCriticalSection(ListCS);
  
finalization
  ClearSections;
  DeleteCriticalSection(ListCS);
  SectionsList.Free;

end.
